<?php
/**
 * Copyright © Lyra Network and contributors.
 * This file is part of Systempay plugin for WooCommerce. See COPYING.md for license details.
 *
 * @author    Lyra Network (https://www.lyra.com/)
 * @author    Geoffrey Crofte, Alsacréations (https://www.alsacreations.fr/)
 * @copyright Lyra Network and contributors
 * @license   http://www.gnu.org/licenses/old-licenses/gpl-2.0.html GNU General Public License (GPL v2)
 */

if (! defined('ABSPATH')) {
    exit; // Exit if accessed directly.
}

class WC_Gateway_SystempaySubscription extends WC_Gateway_SystempayStd
{
    const SUBSCRIPTIONS_HANDLER = 'wc-subscriptions';
    protected $subscriptions_handler;

    public function __construct()
    {
        $this->id = 'systempaysubscription';
        $this->icon = apply_filters('woocommerce_systempaysubscription_icon', WC_SYSTEMPAY_PLUGIN_URL . 'assets/images/systempay.png');
        $this->has_fields = true;
        $this->method_title = self::GATEWAY_NAME. ' - ' . __('Subscription payment', 'woo-systempay-payment');

        $this->supports = array(
            'subscriptions',
            'subscription_cancellation',
            'subscription_payment_method_change',
            'subscription_amount_changes',
            'subscription_date_changes',
            'subscription_payment_method_change_customer',
            'gateway_scheduled_payments'
        );

        // Init common vars.
        $this->systempay_init();

        // Load the form fields.
        $this->init_form_fields();

        // Load the module settings.
        $this->init_settings();

        // Define user set variables.
        $this->title = $this->get_title();
        $this->description = $this->get_description();
        $this->testmode = ($this->get_general_option('ctx_mode') == 'TEST');
        $this->debug = ($this->get_general_option('debug') == 'yes') ? true : false;

        // Use the selected susbscriptions handler.
        $handler = $this->get_option('subscriptions') ? $this->get_option('subscriptions') : self::SUBSCRIPTIONS_HANDLER;
        $this->subscriptions_handler = Systempay_Subscriptions_Loader::getInstance($handler);

        if ($this->systempay_is_section_loaded()) {
            // Reset subscription payment admin form action.
            add_action('woocommerce_settings_start', array($this, 'systempay_reset_admin_options'));

            // Update subscription payment admin form action.
            add_action('woocommerce_update_options_payment_gateways_' . $this->id, array($this, 'process_admin_options'));

            // Adding style to admin form action.
            add_action('admin_head-woocommerce_page_' . $this->admin_page, array($this, 'systempay_admin_head_style'));

            // Adding JS to admin form action.
            add_action('admin_head-woocommerce_page_' . $this->admin_page, array($this, 'systempay_admin_head_script'));
        }

        // Generate subscription payment form action.
        add_action('woocommerce_receipt_' . $this->id, array($this, 'systempay_generate_form'));

        // Iframe payment endpoint action.
        add_action('woocommerce_api_wc_gateway_' . $this->id, array($this, 'systempay_generate_iframe_form'));

        // Order needs payment filter.
        add_filter('woocommerce_order_needs_payment', array($this, 'systempay_order_needs_payment'), 10, 2);

        $this->subscriptions_handler->init_hooks();
    }

    /**
     * Initialise gateway settings form fields.
     */
    public function init_form_fields()
    {
        parent::init_form_fields();

        unset($this->form_fields['payment_cards']);
        unset($this->form_fields['card_data_mode']);
        unset($this->form_fields['payment_by_token']);

        // By default, disable Subscription payment submodule.
        $this->form_fields['enabled']['default'] = 'no';
        $this->form_fields['enabled']['description'] = __('Enables / disables payment by Subscription.', 'woo-systempay-payment');

        // Add subscription payment fields.
        $this->form_fields['subscriptions'] = array(
            'title' => __('Subscriptions management', 'woo-systempay-payment'),
            'type' => 'select',
            'default' => 'wc-subscriptions',
            'options' => array(
                'wc-subscriptions' => 'WooCommerce Subscriptions',
                'subscriptio' => 'Subscriptio 2.x',
                'custom' => __('Custom', 'woo-systempay-payment')
            ),
            'description' => __('If you buy subscriptions on your site, choose the solution you use to manage them. If you choose "Custom", your developper may develop a subscriptions adapter for our plugin.', 'woo-systempay-payment'),
            'class' => 'wc-enhanced-select'
       );
    }

    protected function get_rest_fields()
    {
        // REST API fields are not available for this payment.
    }

    protected function is_available_for_subscriptions()
    {
        global $woocommerce;

        $cust_id = $this->get_customer_property($woocommerce->customer, 'id');
        if (! $cust_id) {
            return false;
        }

        if ($this->subscriptions_handler
            && ($this->subscriptions_handler->cart_contains_multiple_subscriptions($woocommerce->cart)
                || ! $this->subscriptions_handler->cart_contains_subscription($woocommerce->cart))) {
            return false;
        }

        return true;
    }

    protected function payment_by_alias_view($html, $force_redir = true)
    {
        global $woocommerce;

        $cust_id = $this->get_customer_property($woocommerce->customer, 'id');
        $saved_subsc_masked_pan = get_user_meta((int) $cust_id, $this->id . '_masked_pan', true);

        // Recover card brand if saved with masked pan and check if logo exists.
        $card_brand = '';
        $card_brand_logo = '';
        if (strpos($saved_subsc_masked_pan, '|')) {
            $card_brand = substr($saved_subsc_masked_pan, 0, strpos($saved_subsc_masked_pan, '|'));
            $remote_logo = self::LOGO_URL . strtolower($card_brand) . '.png';
            if ($card_brand) {
                $card_brand_logo = '<img src="' . $remote_logo . '"
                       alt="' . $card_brand . '"
                       title="' . $card_brand . '"
                       style="vertical-align: middle; margin: 0 10px 0 5px; max-height: 20px; display: unset;">';
            }
        }

        $saved_subsc_masked_pan = $card_brand_logo ? $card_brand_logo . '<b style="vertical-align: middle;">' . substr($saved_subsc_masked_pan, strpos($saved_subsc_masked_pan, '|') + 1) . '</b>'
            : ' <b>' . str_replace('|',' ', $saved_subsc_masked_pan) . '</b>';

        echo '<div id="' . $this->id . '_payment_by_token_description">
                  <ul>
                      <li>
                          <span>' .
                              sprintf(__('You will pay with your stored means of payment %s', 'woo-systempay-payment'), $saved_subsc_masked_pan)
                              . ' (<a href="' . esc_url(wc_get_account_endpoint_url($this->get_option('woocommerce_saved_cards_endpoint', 'ly_saved_cards'))) . '">' . __('manage your payment means', 'woo-systempay-payment') . '</a>).
                          </span>
                      </li>
                  </ul>
              </div>';
    }

    protected function can_use_alias($cust_id, $verify_identifier = false)
    {
        global $woocommerce;

        if (! $cust_id) {
            return false;
        }

        $amount = $woocommerce->cart ? $woocommerce->cart->total : 0;
        if (($amount <= 0) && (! $verify_identifier || (! empty($_GET['wc-ajax']) && $this->check_identifier($cust_id, $this->id)))) {
            return true;
        }

        return false;
    }

    /**
     * Prepare form params to send to payment gateway.
     **/
    protected function systempay_fill_request($order)
    {
        parent::systempay_fill_request($order);

        $info = $this->subscriptions_handler->subscription_info($order);

        if (is_array($info) && ! empty($info)) {
            $currency = SystempayApi::findCurrencyByAlphaCode(get_woocommerce_currency());
            $cust_id = $this->get_order_property($order, 'user_id');

            $this->systempay_request->set('sub_amount', $currency->convertAmountToInteger($info['amount']));
            $this->systempay_request->set('sub_currency', $currency->getNum()); // Same as general order currency.

            $desc = 'RRULE:FREQ=' . $info['frequency']. ';INTERVAL=' . $info['interval'];
            if (isset($info['end_date']) && $info['end_date']) {
                $desc .= ';UNTIL=' . $info['end_date'];
            }

            $this->systempay_request->set('sub_desc', $desc);

            $this->systempay_request->set('sub_effect_date', $info['effect_date']);

            // Initial amount.
            if (isset($info['init_amount']) && $info['init_amount'] && isset($info['init_number']) && $info['init_number']) {
                $this->systempay_request->set('sub_init_amount', $currency->convertAmountToInteger($info['init_amount']));
                $this->systempay_request->set('sub_init_amount_number', $info['init_number']);
            }

            if ($order->get_total() > 0) {
                $this->systempay_request->set('page_action', 'REGISTER_PAY_SUBSCRIBE');
            } else {
                // Only subscriptions.
                $saved_identifier = $this->get_cust_identifier($cust_id);
                $is_identifier_active = $this->is_cust_identifier_active($cust_id);
                if ($saved_identifier && $is_identifier_active) {
                    $this->systempay_request->set('identifier', $saved_identifier);
                    $this->systempay_request->set('page_action', 'SUBSCRIBE');
                } else {
                    $this->systempay_request->set('page_action', 'REGISTER_SUBSCRIBE');
                }
            }
        }
    }

    public function cancel_online_subscription($subscription_id, $order_id)
    {
        $key = $this->testmode ? $this->get_general_option('test_private_key') : $this->get_general_option('prod_private_key');
        if (! $key) {
            $this->log("Subscription #{$subscription_id} cannot be cancelled on gateway for order #$order_id: private key is not configured.");
            return;
        }

        $this->log("Cancelling subscription #{$subscription_id} for order #$order_id.");

        $order = new WC_Order((int) $order_id);

        $cust_id = $this->get_order_property($order, 'user_id');
        $saved_identifier = $this->get_cust_identifier($cust_id);

        $params = array(
            'subscriptionId' => get_post_meta($order_id, 'Subscription ID', true),
            'paymentMethodToken' => $saved_identifier
        );

        try {
            $client = new SystempayRest(
                $this->get_general_option('rest_url'),
                $this->get_general_option('site_id'),
                $key
            );

            $result = $client->post('V4/Subscription/Cancel', json_encode($params));
            SystempayRestTools::checkResult($result);

            //Subscription cancelled successfully.
            $this->log("Subscription #{$subscription_id} cancelled successfully for order #$order_id.");

            // Cancel running subscriptions.
            $transactions = $this->get_order_details($order_id);

            foreach ($transactions as $transaction) {
                $this->cancel_transaction($transaction);
            }
        } catch (Exception $e) {
            $this->log("Subscription cancel exception for order #$order_id with code {$e->getCode()}: {$e->getMessage()}.");
        }
    }

    private function cancel_transaction($transaction)
    {
        if ($transaction['status'] !== 'RUNNING' ){
            return;
        }

        $order_id =  $transaction['orderDetails']['orderId'];

        $this->log("Cancelling transaction with UUID #{$transaction['uuid']} for order #$order_id.");

        $key = $this->testmode ? $this->get_general_option('test_private_key') : $this->get_general_option('prod_private_key');
        if (! $key) {
            $this->log("Transaction with UUID #{$transaction['uuid']} cannot be cancelled on gateway for order #$order_id: private key is not configured.");
            return;
        }

        try {
            $params = array(
                'uuid' => $transaction['uuid'],
                'resolutionMode' => 'CANCELLATION_ONLY'
            );

            $client = new SystempayRest(
                $this->get_general_option('rest_url'),
                $this->get_general_option('site_id'),
                $key
            );

            $result = $client->post('V4/Transaction/Cancel', json_encode($params));
            SystempayRestTools::checkResult($result, 'CANCELLED');
        } catch (Exception $e) {
            $this->log("Transaction cancel exception for order #$order_id with code {$e->getCode()}: {$e->getMessage()}.");
        }
    }

    public function update_online_subscription($subscription_id, $order_id)
    {
        $key = $this->testmode ? $this->get_general_option('test_private_key') : $this->get_general_option('prod_private_key');
        if (! $key) {
            $this->log("Subscription #{$subscription_id} for order #$order_id cannot be updated on gateway: private key is not configured.");
            return;
        }

        $this->log("Updating subscription #{$subscription_id} for order #$order_id.");

        $order = new WC_Order((int) $order_id);

        $cust_id = $this->get_order_property($order, 'user_id');
        $saved_identifier = $this->get_cust_identifier($cust_id);

        // Re-generate subscriptions data from updated order.
        $info = $this->subscriptions_handler->subscription_info($order);
        if (! is_array($info) || empty($info)) {
            $this->log("Empty subscription info returned for order #$order_id. Cannot update subscription #{$subscription_id}.");
            return;
        }

        // Get subscription effect date in ISO8601 format.
        $date_time = strtotime($info['effect_date']);
        if ($date_time < time()) {
            $this->log("Cannot update subscription #{$subscription_id} for order #$order_id. Effect date has passed.");
            return;
        }

        // Compute new rrule.
        $desc = 'RRULE:FREQ=' . $info['frequency']. ';INTERVAL=' . $info['interval'];
        if (isset($info['end_date']) && $info['end_date']) {
            $desc .= ';UNTIL=' . $info['end_date'];
        }

        $currency = SystempayApi::findCurrencyByAlphaCode(get_woocommerce_currency());

        $params = array(
            'subscriptionId' => get_post_meta($order_id, 'Subscription ID', true),
            'paymentMethodToken' => $saved_identifier,
            'amount' => $currency->convertAmountToInteger($info['amount']),
            'currency' => $currency->getAlpha3(),
            'effectDate' => date(DateTime::ATOM, $date_time),
            'rrule' => $desc
        );

        // Initial amount.
        if (isset($info['init_amount']) && $info['init_amount'] && isset($info['init_number']) && $info['init_number']) {
            $params['initialAmount'] = $currency->convertAmountToInteger($info['init_amount']);
            $params['initialAmountNumber'] = $info['init_number'];
        }

        try {
            $client = new SystempayRest(
                $this->get_general_option('rest_url'),
                $this->get_general_option('site_id'),
                $key
            );

            $result = $client->post('V4/Subscription/Update', json_encode($params));
            SystempayRestTools::checkResult($result);

            //Subscription cancelled successfully.
            $this->log("Subscription #{$subscription_id} updated successfully for order #$order_id.");
        } catch (Exception $e) {
            $this->log("Subscription update exception for order #$order_id with code {$e->getCode()}: {$e->getMessage()}");
        }
    }

    private function get_order_details($order_id)
    {
        $key = $this->testmode ? $this->get_general_option('test_private_key') : $this->get_general_option('prod_private_key');
        $client = new SystempayRest(
            $this->get_general_option('rest_url'),
            $this->get_general_option('site_id'),
            $key
        );

        $params = array(
            'orderId' => $order_id,
            'operationType' => 'DEBIT'
        );

        $get_order_response = $client->post('V4/Order/Get', json_encode($params));
        SystempayRestTools::checkResult($get_order_response);

        // Order transactions organized by sequence numbers.
        $trans_by_sequence = array();
        foreach ($get_order_response['answer']['transactions'] as $transaction) {
            $sequence_number = $transaction['transactionDetails']['sequenceNumber'];
            // Unpaid transactions are not considered.
            if ($transaction['status'] !== 'UNPAID') {
                $trans_by_sequence[$sequence_number] = $transaction;
            }
        }

        ksort($trans_by_sequence);
        return array_reverse($trans_by_sequence);
    }

    public function systempay_order_needs_payment($is_active, $order)
    {
        global $woocommerce;

        if (($order->get_total() == 0) && ($this->get_order_property($order, 'payment_method') === $this->id)) {
            return $this->subscriptions_handler && $this->subscriptions_handler->cart_contains_subscription($woocommerce->cart);
        }

        return $is_active;
    }
}
